@@ -10,6 +10,9 @@
 """
 XSD atomic datatypes for XPath. Includes a class for UntypedAtomic data and some
 classes for XSD datetime and duration types.
+
+TODO for v1.5: Create special wrapper classes for xs:double and other numerical
+    types for the properly working of 'instance of' tests.
 """
 from abc import ABCMeta, abstractmethod
 import operator
@@ -184,6 +187,7 @@ class AbstractDateTime(metaclass=ABCMeta):
     version = '1.0'
     _pattern = re.compile(r'^$')
     _utc_timezone = Timezone(datetime.timedelta(0))
+    _year = None
 
     def __init__(self, year=2000, month=1, day=1, hour=0, minute=0,
                  second=0, microsecond=0, tzinfo=None):
@@ -194,7 +198,6 @@ class AbstractDateTime(metaclass=ABCMeta):
             delta = 0
 
         if 1 <= year <= 9999:
-            self._year = None
             self._dt = datetime.datetime(year, month, day, hour, minute,
                                          second, microsecond, tzinfo)
         elif year == 0:
@@ -231,10 +234,7 @@ class AbstractDateTime(metaclass=ABCMeta):
 
     @abstractmethod
     def __str__(self):
-        pass
-
-    def __unicode__(self):
-        return str(self)
+        raise NotImplementedError()
 
     @property
     def year(self):
@@ -251,7 +251,7 @@ class AbstractDateTime(metaclass=ABCMeta):
         if -9999 <= year < -1:
             return '{:05}'.format(year if self.version == '1.0' else year + 1)
         elif year == -1:
-            return '{:04}'.format(year if self.version == '1.0' else year + 1)
+            return '-0001' if self.version == '1.0' else '0000'
         elif 0 <= year <= 9999:
             return '{:04}'.format(year)
         else:
@@ -358,7 +358,7 @@ class AbstractDateTime(metaclass=ABCMeta):
             else:
                 return self._dt, dt
         else:
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
 
     def __hash__(self):
         return hash((self._dt, self._year))
@@ -366,20 +366,20 @@ class AbstractDateTime(metaclass=ABCMeta):
     def __eq__(self, other):
         try:
             return operator.eq(*self._get_operands(other)) and self.year == other.year
-        except ElementPathTypeError:
+        except TypeError:
             return False
 
     def __ne__(self, other):
         try:
             return operator.ne(*self._get_operands(other)) or self.year != other.year
-        except ElementPathTypeError:
+        except TypeError:
             return True
 
 
 class OrderedDateTime(AbstractDateTime):
     @abstractmethod
     def __str__(self):
-        pass
+        raise NotImplementedError()
 
     @classmethod
     def fromdelta(cls, delta, adjust_timezone=False):
@@ -388,7 +388,7 @@ class OrderedDateTime(AbstractDateTime):
         0001-01-01T00:00:00 CE. In case of a date the time part is not counted.
 
         :param delta: a datetime.timedelta instance.
-        :param adjust_timezone: if `True` adjust the timezone of Date objects \
+        :param adjust_timezone: if `True` adjusts the timezone of Date objects \
         with eventually present hours and minutes.
         """
         try:
@@ -437,12 +437,9 @@ class OrderedDateTime(AbstractDateTime):
             year = dt.year
 
         if issubclass(cls, Date10):
-            if adjust_timezone and dt.hour or dt.minute:
-                if dt.tzinfo is None:
-                    hour, minute = dt.hour, dt.minute
-                else:
-                    hour = dt.hour - dt.tzinfo.offset.hours
-                    minute = dt.minute - dt.tzinfo.offset.minutes
+            if adjust_timezone and (dt.hour or dt.minute):
+                assert dt.tzinfo is None
+                hour, minute = dt.hour, dt.minute
 
                 if hour < 14 or hour == 14 and minute == 0:
                     tz = Timezone(datetime.timedelta(hours=-hour, minutes=-minute))
@@ -459,7 +456,7 @@ class OrderedDateTime(AbstractDateTime):
     def todelta(self):
         """Returns the datetime.timedelta from 0001-01-01T00:00:00 CE."""
         if self._year is None:
-            return self._dt - datetime.datetime(1, 1, 1)
+            return operator.sub(*self._get_operands(datetime.datetime(1, 1, 1)))
 
         year, dt = self.year, self._dt
         tzinfo = None if dt.tzinfo is None else self._utc_timezone
@@ -481,40 +478,33 @@ class OrderedDateTime(AbstractDateTime):
                 return DayTimeDuration.fromtimedelta(dt1 - dt2)
             return DayTimeDuration.fromtimedelta(self.todelta() - other.todelta())
 
-        elif isinstance(other, (DayTimeDuration, datetime.timedelta)):
-            delta = other.get_timedelta() if isinstance(other, DayTimeDuration) else other
-            try:
-                dt = op(self._dt, delta)
-            except OverflowError:
-                seconds = delta.total_seconds()
-                if self.bce ^ (seconds > 0):
-                    raise
-                dt_seconds = abs((self._dt - datetime.datetime(1, 1, 1)).total_seconds() - seconds)
-                dt = datetime.datetime(1, 1, 1) + datetime.timedelta(seconds=dt_seconds)
-                bce = not self.bce
-            else:
-                bce = self.bce
+        elif isinstance(other, datetime.timedelta):
+            delta = op(self.todelta(), other)
+            return type(self).fromdelta(delta, adjust_timezone=True)
 
-            if 'hour' not in self._pattern.groupindex.keys():
-                dt = dt.replace(hour=0, minute=0, second=0, microsecond=0)
+        elif isinstance(other, DayTimeDuration):
+            delta = op(self.todelta(), other.get_timedelta())
+            return type(self).fromdelta(delta)
 
         elif isinstance(other, YearMonthDuration):
             month = op(self._dt.month - 1, other.months) % 12 + 1
-            year = op(self._dt.year, ((self._dt.month - 1) + other.months) // 12)
+            year = op(self.year, ((self._dt.month - 1) + other.months) // 12)
+            day = adjust_day(year, month, self._dt.day)
+
             if year > 0:
-                bce = self.bce
+                dt = self._dt.replace(year=year, month=month, day=day)
+            elif isleap(year):
+                dt = self._dt.replace(year=4, month=month, day=day)
             else:
-                bce = not self.bce
-                year = abs(year) + 1
-            dt = self._dt.replace(year=year, month=month, day=adjust_day(year, month, self._dt.day))
+                dt = self._dt.replace(year=6, month=month, day=day)
+
+            kwargs = {k: getattr(dt, k) for k in self._pattern.groupindex.keys()}
+            if year <= 0:
+                kwargs['year'] = year
+            return type(self)(**kwargs)
 
         else:
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
-
-        kwargs = {k: getattr(dt, k) for k in self._pattern.groupindex.keys()}
-        if bce:
-            kwargs['year'] = -kwargs['year']
-        return type(self)(**kwargs)
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
 
     def __lt__(self, other):
         dt1, dt2 = self._get_operands(other)
@@ -538,7 +528,7 @@ class OrderedDateTime(AbstractDateTime):
 
     def __add__(self, other):
         if isinstance(other, OrderedDateTime):
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
         return self._date_operator(operator.add, other)
 
     def __sub__(self, other):
@@ -725,7 +715,7 @@ class Time(AbstractDateTime):
         elif isinstance(other, datetime.timedelta):
             dt = self._dt + other
         else:
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
         return Time(dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)
 
     def __sub__(self, other):
@@ -735,8 +725,11 @@ class Time(AbstractDateTime):
         elif isinstance(other, DayTimeDuration):
             dt = self._dt - other.get_timedelta()
             return Time(dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)
+        elif isinstance(other, datetime.timedelta):
+            dt = self._dt - other
+            return Time(dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)
         else:
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
 
 
 class Duration(object):
@@ -793,9 +786,6 @@ class Duration(object):
             value += 'T0S'
         return value
 
-    def __unicode__(self):
-        return str(self)
-
     @classmethod
     def fromstring(cls, text):
         """
@@ -806,7 +796,7 @@ class Duration(object):
         """
         match = cls._pattern.match(text)
         if match is None:
-            raise ElementPathValueError('%r is not an xs:duration value.' % text)
+            raise ElementPathValueError('%r is not an xs:duration value' % text)
 
         sign, years, months, days, hours, minutes, seconds = match.groups()
         seconds = decimal.Decimal(seconds or 0)
@@ -826,11 +816,11 @@ class Duration(object):
 
         if cls is DayTimeDuration:
             if months:
-                raise ElementPathValueError('months must be 0 for %r.' % cls.__name__)
+                raise ElementPathValueError('months must be 0 for %r' % cls.__name__)
             return cls(seconds=seconds)
         elif cls is YearMonthDuration:
             if seconds:
-                raise ElementPathValueError('seconds must be 0 for %r.' % cls.__name__)
+                raise ElementPathValueError('seconds must be 0 for %r' % cls.__name__)
             return cls(months=months)
         return cls(months=months, seconds=seconds)
 
@@ -844,15 +834,9 @@ class Duration(object):
 
         Ref: https://www.w3.org/TR/2012/REC-xmlschema11-2-20120405/#duration
         """
-        if isinstance(other, self.__class__):
-            pass
-        elif other.__class__ is Duration:
-            if isinstance(self.__class__, DayTimeDuration) and other.months:
-                raise ElementPathValueError("the 2nd operand must has 0 months value")
-            elif isinstance(self.__class__, YearMonthDuration) and other.months:
-                raise ElementPathValueError("the 2nd operand must has 0 days value")
-        else:
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+        if not isinstance(other, self.__class__):
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
+
         m1, s1 = self.months, int(self.seconds)
         m2, s2 = other.months, int(other.seconds)
         ms1, ms2 = int((self.seconds - s1) * 1000000), int((other.seconds - s2) * 1000000)
@@ -905,17 +889,17 @@ class YearMonthDuration(Duration):
 
     def __add__(self, other):
         if not isinstance(other, self.__class__):
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
         return YearMonthDuration(months=self.months + other.months)
 
     def __sub__(self, other):
         if not isinstance(other, self.__class__):
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
         return YearMonthDuration(months=self.months - other.months)
 
     def __mul__(self, other):
         if not isinstance(other, (float, int, decimal.Decimal)):
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
         return YearMonthDuration(months=int(float(self.months * other) + 0.5))
 
     def __truediv__(self, other):
@@ -924,7 +908,7 @@ class YearMonthDuration(Duration):
         elif isinstance(other, (float, int, decimal.Decimal)):
             return YearMonthDuration(months=int(float(self.months / other) + 0.5))
         else:
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
 
 
 class DayTimeDuration(Duration):
@@ -948,17 +932,17 @@ class DayTimeDuration(Duration):
 
     def __add__(self, other):
         if not isinstance(other, self.__class__):
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
         return DayTimeDuration(seconds=self.seconds + other.seconds)
 
     def __sub__(self, other):
         if not isinstance(other, self.__class__):
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
         return DayTimeDuration(seconds=self.seconds - other.seconds)
 
     def __mul__(self, other):
         if not isinstance(other, (float, int, decimal.Decimal)):
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
         return DayTimeDuration(seconds=int(float(self.seconds * other) + 0.5))
 
     def __truediv__(self, other):
@@ -967,7 +951,7 @@ class DayTimeDuration(Duration):
         elif isinstance(other, (float, int, decimal.Decimal)):
             return DayTimeDuration(seconds=int(float(self.seconds / other) + 0.5))
         else:
-            raise ElementPathTypeError("wrong type %r for operand %r." % (type(other), other))
+            raise ElementPathTypeError("wrong type %r for operand %r" % (type(other), other))
 
 
 class UntypedAtomic(object):
@@ -978,10 +962,19 @@ class UntypedAtomic(object):
     :param value: the untyped value, usually a string.
     """
     def __init__(self, value):
-        self.value = value
+        if not isinstance(value, (str, bytes, int, float, decimal.Decimal)):
+            raise ElementPathTypeError("{!r} is not an atomic value".format(value))
+        elif isinstance(value, str):
+            self.value = value
+        elif isinstance(value, bytes):
+            self.value = value.decode('utf-8')
+        elif isinstance(value, bool):
+            self.value = 'true' if value else 'false'
+        else:
+            self.value = str(value)
 
     def __repr__(self):
-        return '%s(value=%r)' % (self.__class__.__name__, self.value)
+        return '%s(%r)' % (self.__class__.__name__, self.value)
 
     def _get_operands(self, other, force_float=True):
         """
@@ -995,8 +988,9 @@ class UntypedAtomic(object):
         if isinstance(other, UntypedAtomic):
             if force_float:
                 return float(self.value), float(other.value)
-            else:
-                return self.value, other.value
+            return self.value, other.value
+        elif isinstance(other, bool):
+            return bool(self), other
         elif isinstance(other, int):
             return float(self.value), other
         else:
@@ -1050,16 +1044,19 @@ class UntypedAtomic(object):
         return float(self.value)
 
     def __bool__(self):
-        return bool(self.value)
+        value = self.value.strip()
+        if value not in {'0', '1', 'true', 'false'}:
+            raise ElementPathValueError("{!r} cannot be cast to boolean".format(self.value))
+        return value in ('1', 'true')
 
     def __abs__(self):
-        return abs(self.value)
+        return abs(decimal.Decimal(self.value))
 
     def __mod__(self, other):
         return operator.mod(*self._get_operands(other))
 
     def __str__(self):
-        return str(self.value)
+        return self.value
 
     def __bytes__(self):
         return bytes(self.value, encoding='utf-8')
@@ -1169,7 +1166,9 @@ def integer_validator(x):
 
 
 def base64_binary_validator(x):
-    if not isinstance(x, str) or NOT_BASE64_BINARY_PATTERN.match(x) is None:
+    if isinstance(x, bytes):
+        x = x.decode()
+    if not isinstance(x, str) or NOT_BASE64_BINARY_PATTERN.search(x) is not None:
         return False
     try:
         base64.standard_b64decode(x)
@@ -1187,7 +1186,15 @@ def ncname_validator(x):
     return isinstance(x, str) and NCNAME_PATTERN.match(x) is not None
 
 
-XSD_BUILTIN_TYPES = {
+def is_id(obj):
+    return isinstance(obj, str) and ncname_validator(obj)
+
+
+def is_idrefs(obj):
+    return isinstance(obj, str) and all(ncname_validator(x) for x in obj.split())
+
+
+XSD_BUILTIN_TYPES = {           # pragma: no cover
     'anyType': XsdBuiltin(
         lambda x: True,
         value=UntypedAtomic('1')
