@@ -9,6 +9,7 @@
 #
 import math
 import decimal
+from copy import copy
 
 from .exceptions import ElementPathSyntaxError, ElementPathNameError, MissingContextError
 from .datatypes import UntypedAtomic, DayTimeDuration, YearMonthDuration, \
@@ -85,7 +86,7 @@ class XPath1Parser(Parser):
     # Labels and symbols admitted after a path step
     PATH_STEP_LABELS = ('axis', 'kind test')
     PATH_STEP_SYMBOLS = {
-        '(integer)', '(string)', '(float)', '(decimal)', '(name)', '*', '@', '..', '.', '(', '{'
+        '(integer)', '(string)', '(float)', '(decimal)', '(name)', '*', '@', '..', '.', '{'
     }
 
     schema = None  # To simplify the schema bind checks in compatibility with XPath2Parser
@@ -246,31 +247,30 @@ def select(self, context=None):
         return
 
     name = self.value
+
     if isinstance(context, XPathSchemaContext):
         # Bind with the XSD type from a schema
-        for schema_item in context.iter_children_or_self():
-            if self.match_xsd_type(schema_item, name) is not None:
-                yield self.get_typed_node(context, schema_item)
+        for item in map(lambda x: self.match_xsd_type(x, name), context.iter_children_or_self()):
+            if item:
+                yield item
         return
 
-    if name[0] != '{' and self.parser.default_namespace:
-        tag = '{%s}%s' % (self.parser.default_namespace, name)
-    else:
+    if name[0] == '{' or not self.parser.default_namespace:
         tag = name
+    else:
+        tag = '{%s}%s' % (self.parser.default_namespace, name)
 
-    # Checks if the token is bound to an XSD type. If not try a match using
-    # the element path. If this match fails the xsd_type attribute is set
-    # with the schema object to prevent other checks until the schema change.
-    if self.xsd_type is self.parser.schema:
+    # With an ElementTree context checks if the token is bound to an XSD type. If not
+    # try a match using the element path. If this match fails the xsd_type attribute
+    # is set with the schema object to prevent other checks until the schema change.
+    if self.xsd_types is self.parser.schema:
 
         # Untyped selection
         for item in context.iter_children_or_self():
-            if is_attribute_node(item, name):
-                yield item
-            elif is_element_node(item, tag):
+            if is_attribute_node(item, name) or is_element_node(item, tag):
                 yield item
 
-    elif self.xsd_type is None or isinstance(self.xsd_type, AbstractSchemaProxy):
+    elif self.xsd_types is None or isinstance(self.xsd_types, AbstractSchemaProxy):
 
         # Try to match the type using the path
         for item in context.iter_children_or_self():
@@ -279,16 +279,16 @@ def select(self, context=None):
 
                 xsd_component = self.parser.schema.find(path, self.parser.namespaces)
                 if xsd_component is not None:
-                    self.xsd_type = xsd_component.type
+                    self.xsd_types = {tag: xsd_component.type}
                 else:
-                    self.xsd_type = self.parser.schema
+                    self.xsd_types = self.parser.schema
 
-                yield self.get_typed_node(context, item)
+                yield self.get_typed_node(item)
     else:
         # XSD typed selection
         for item in context.iter_children_or_self():
             if is_attribute_node(item, name) or is_element_node(item, tag):
-                yield self.get_typed_node(context, item)
+                yield self.get_typed_node(item)
 
 
 ###
@@ -310,6 +310,7 @@ def led(self, left):
     if self.parser.is_spaced():
         self.wrong_syntax("a QName cannot contains spaces before or after ':'")
     self[:] = left, self.parser.expression(90)
+    self.value = '{}:{}'.format(self[0].value, self[1].value)
     return self
 
 
@@ -329,42 +330,41 @@ def select(self, context=None):
         else:
             yield value
         return
-    elif self[0].value == '*':
-        value = '*:%s' % self[1].value
+
+    if self[0].value == '*':
+        name = '*:%s' % self[1].value
     else:
         namespace = self.get_namespace(self[0].value)
-        value = '{%s}%s' % (namespace, self[1].value)
+        name = '{%s}%s' % (namespace, self[1].value)
 
     if context is None:
         return
     elif isinstance(context, XPathSchemaContext):
-        for schema_item in context.iter_children_or_self():
-            if self.match_xsd_type(schema_item, value) is not None:
-                yield self.get_typed_node(context, schema_item)
-
-    elif self.xsd_type is self.parser.schema:
-        for item in context.iter_children_or_self():
-            if is_attribute_node(item, value):
-                yield item
-            elif is_element_node(item, value):
+        for item in map(lambda x: self.match_xsd_type(x, name), context.iter_children_or_self()):
+            if item:
                 yield item
 
-    elif self.xsd_type is None or isinstance(self.xsd_type, AbstractSchemaProxy):
+    elif self.xsd_types is self.parser.schema:
         for item in context.iter_children_or_self():
-            if is_attribute_node(item, value) or is_element_node(item, value):
+            if is_attribute_node(item, name) or is_element_node(item, name):
+                yield item
+
+    elif self.xsd_types is None or isinstance(self.xsd_types, AbstractSchemaProxy):
+        for item in context.iter_children_or_self():
+            if is_attribute_node(item, name) or is_element_node(item, name):
                 path = context.get_path(item)
                 xsd_component = self.parser.schema.find(path, self.parser.namespaces)
                 if xsd_component is not None:
-                    self.xsd_type = xsd_component.type
+                    self.add_xsd_type(xsd_component.name, xsd_component.type)
                 else:
-                    self.xsd_type = self.parser.schema
-                yield self.get_typed_node(context, item)
+                    self.xsd_types = self.parser.schema
+                yield self.get_typed_node(item)
 
     else:
         # XSD typed selection
         for item in context.iter_children_or_self():
-            if is_attribute_node(item, value) or is_element_node(item, value):
-                yield self.get_typed_node(context, item)
+            if is_attribute_node(item, name) or is_element_node(item, name):
+                yield self.get_typed_node(item)
 
 
 ###
@@ -409,8 +409,8 @@ def select(self, context=None):
 def nud(self):
     self.parser.next_token.expected('(name)')
     self[:] = self.parser.expression(rbp=90),
-    if self[0].value.startswith(':'):
-        self[0].wrong_syntax("Variable reference requires a simple reference name")
+    if ':' in self[0].value:
+        self[0].wrong_syntax("variable reference requires a simple reference name")
     return self
 
 
@@ -445,6 +445,8 @@ def select(self, context=None):
         # Wildcard literal
         for item in context.iter_children_or_self():
             if context.is_principal_node_kind():
+                if hasattr(item, 'type'):
+                    self.add_xsd_type(item.name, item.type)
                 if is_attribute_node(item):
                     yield item[1]
                 else:
@@ -455,8 +457,11 @@ def select(self, context=None):
 def select(self, context=None):
     if context is None:
         self.missing_context()
+
     for item in context.iter_self():
         if item is not None:
+            if hasattr(item, 'type') and isinstance(context, XPathSchemaContext):
+                self.add_xsd_type(item.name, item.type)
             yield item
         elif is_document_node(context.root):
             yield context.root
@@ -477,18 +482,14 @@ def select(self, context=None):
 # Logical Operators
 @method(infix('or', bp=20))
 def evaluate(self, context=None):
-    if context is None:
-        return self.boolean_value(self[0].evaluate()) or self.boolean_value(self[1].evaluate())
-    return self.boolean_value(self[0].evaluate(context.copy())) or \
-        self.boolean_value(self[1].evaluate(context.copy()))
+    return self.boolean_value(self[0].evaluate(copy(context))) or \
+        self.boolean_value(self[1].evaluate(copy(context)))
 
 
 @method(infix('and', bp=25))
 def evaluate(self, context=None):
-    if context is None:
-        return self.boolean_value(self[0].evaluate()) and self.boolean_value(self[1].evaluate())
-    return self.boolean_value(self[0].evaluate(context.copy())) and \
-        self.boolean_value(self[1].evaluate(context.copy()))
+    return self.boolean_value(self[0].evaluate(copy(context))) and \
+        self.boolean_value(self[1].evaluate(copy(context)))
 
 
 @method(infix('=', bp=30))
@@ -566,6 +567,9 @@ def evaluate(self, context=None):
                 return op1 * op2
             except TypeError as err:
                 raise self.wrong_type(str(err))
+    else:
+        # This is not a multiplication operator but a wildcard select statement
+        return [x for x in self.select(context)]
 
 
 @method(infix('div', bp=45))
@@ -660,13 +664,12 @@ def select(self, context=None):
         if is_document_node(context.root):
             yield context.root
     elif len(self) == 1:
-        context.item = None
-        yield from self[0].select(context)
+        if is_document_node(context.root) or context.item is context.root:
+            context.item = None
+            yield from self[0].select(context)
     else:
         items = []
-        left_results = [x for x in self[0].select(context)]
-        context.size = len(left_results)
-        for context.position, context.item in enumerate(left_results):
+        for _ in context.iter_selector(self[0].select):
             if not is_xpath_node(context.item):
                 self.wrong_type("left operand must returns XPath nodes: {}".format(context.item))
             for result in self[1].select(context):
@@ -681,6 +684,11 @@ def select(self, context=None):
                 else:
                     items.append(result)
                     yield result
+                    if isinstance(context, XPathSchemaContext):
+                        if isinstance(result, tuple):
+                            self[1].add_xsd_type(result[0], result[1].type)
+                        elif hasattr(result, 'type'):
+                            self[1].add_xsd_type(result.tag, result.type)
 
 
 @method('//')
@@ -688,9 +696,10 @@ def select(self, context=None):
     if context is None:
         return
     elif len(self) == 1:
-        context.item = None
-        for _ in context.iter_descendants(axis='descendant-or-self'):
-            yield from self[0].select(context)
+        if is_document_node(context.root) or context.item is context.root:
+            context.item = None
+            for _ in context.iter_descendants(axis='descendant-or-self'):
+                yield from self[0].select(context)
     else:
         for elem in self[0].select(context):
             if not is_element_node(elem):
@@ -712,13 +721,18 @@ def led(self, left):
 @method('[')
 def select(self, context=None):
     if context is not None:
-        for position, item in enumerate(self[0].select(context), start=1):
+        if self[0].label == 'axis':
+            selector = self[0].select(context)
+        else:
+            selector = context.iter_selector(self[0].select)
+
+        for _ in selector:
             predicate = [x for x in self[1].select(context.copy())]
             if len(predicate) == 1 and isinstance(predicate[0], NumericTypeProxy):
-                if position == predicate[0]:
-                    yield item
+                if context.position == predicate[0]:
+                    yield context.item
             elif self.boolean_value(predicate):
-                yield item
+                yield context.item
 
 
 ###
@@ -728,7 +742,7 @@ def nud(self):
     self.parser.next_token.unexpected(')')
     self[:] = self.parser.expression(),
     self.parser.advance(')')
-    return self  # Skip self!! (remove a redundant level from selection/evaluation)
+    return self
 
 
 @method('(')
@@ -742,75 +756,7 @@ def select(self, context=None):
 
 
 ###
-# Forward Axes
-@method(axis('self'))
-def select(self, context=None):
-    if context is not None:
-        for _ in context.iter_self():
-            yield from self[0].select(context)
-
-
-@method(axis('child'))
-def select(self, context=None):
-    if context is not None:
-        for _ in context.iter_children_or_self(child_axis=True):
-            yield from self[0].select(context)
-
-
-@method(axis('descendant'))
-def select(self, context=None):
-    if context is not None:
-        item = context.item
-        for _ in context.iter_descendants(axis=self.symbol):
-            if item is not context.item:
-                yield from self[0].select(context)
-
-
-@method(axis('descendant-or-self'))
-def select(self, context=None):
-    if context is not None:
-        for _ in context.iter_descendants(axis=self.symbol):
-            yield from self[0].select(context)
-
-
-@method(axis('following-sibling'))
-def select(self, context=None):
-    if context is not None:
-        if isinstance(context.item, TypedElement):
-            item = context.item[0]
-        elif is_etree_element(context.item):
-            item = context.item
-        else:
-            return
-
-        for parent in context.iter_parent(axis=self.symbol):
-            if isinstance(context, XPathSchemaContext):
-                for _ in context.iter_children_or_self(parent, child_axis=True):
-                    yield from self[0].select(context)
-
-            else:
-                follows = False
-                for child in context.iter_children_or_self(parent, child_axis=True):
-                    if follows:
-                        yield from self[0].select(context)
-                    elif item is child:
-                        follows = True
-
-
-@method(axis('following'))
-def select(self, context=None):
-    if context is not None:
-        descendants = set(context.iter_descendants(axis=self.symbol))
-        item = context.item
-        follows = False
-        for elem in context.iter_descendants(item=context.root, axis=self.symbol):
-            if follows:
-                if elem not in descendants:
-                    yield from self[0].select(context)
-            elif item is elem:
-                follows = True
-
-
+# Axes
 @method('@', bp=80)
 def nud(self):
     self[:] = self.parser.expression(rbp=80),
@@ -841,55 +787,63 @@ def select(self, context=None):
 
         if hasattr(elem, 'nsmap'):
             # Add element's namespaces for lxml (and use None for default namespace)
+            # noinspection PyUnresolvedReferences
             for prefix_, uri in elem.nsmap.items():
                 if prefix_ not in namespaces:
                     context.item = NamespaceNode(prefix_, uri)
                     yield context.item
 
 
-###
-# Reverse Axes
+@method(axis('self'))
+def select(self, context=None):
+    if context is not None:
+        for _ in context.iter_self():
+            yield from self[0].select(context)
+
+
+@method(axis('child'))
+def select(self, context=None):
+    if context is not None:
+        for _ in context.iter_children_or_self(child_axis=True):
+            yield from self[0].select(context)
+
+
 @method(axis('parent'))
 def select(self, context=None):
     if context is not None:
-        for _ in context.iter_parent(axis=self.symbol):
+        for _ in context.iter_parent():
+            yield from self[0].select(context)
+
+
+@method(axis('following-sibling'))
+@method(axis('preceding-sibling'))
+def select(self, context=None):
+    if context is not None:
+        for _ in context.iter_siblings(axis=self.symbol):
             yield from self[0].select(context)
 
 
 @method(axis('ancestor'))
-def select(self, context=None):
-    if context is not None:
-        results = [
-            item
-            for _ in context.iter_ancestors(axis=self.symbol)
-            for item in self[0].select(context)
-        ]
-        for result in reversed(results):
-            context.item = result
-            yield result
-
-
 @method(axis('ancestor-or-self'))
 def select(self, context=None):
     if context is not None:
-        item = context.item
-        for elem in reversed([x for x in context.iter_ancestors(axis=self.symbol)]):
-            context.item = elem
-            yield elem
-        yield item
+        for _ in context.iter_ancestors(axis=self.symbol):
+            yield from self[0].select(context)
 
 
-@method(axis('preceding-sibling'))
+@method(axis('descendant'))
+@method(axis('descendant-or-self'))
 def select(self, context=None):
-    if context is not None and is_element_node(context.item):
-        item = context.item
-        for parent in context.iter_parent(axis=self.symbol):
-            for child in parent:
-                if child is item:
-                    break
-                else:
-                    context.item = child
-                    yield from self[0].select(context)
+    if context is not None:
+        for _ in context.iter_descendants(axis=self.symbol):
+            yield from self[0].select(context)
+
+
+@method(axis('following'))
+def select(self, context=None):
+    if context is not None:
+        for _ in context.iter_followings():
+            yield from self[0].select(context)
 
 
 @method(axis('preceding'))
@@ -940,7 +894,7 @@ def evaluate(self, context=None):
 
 @method(function('position', nargs=0))
 def evaluate(self, context=None):
-    return context.position + 1 if context is not None else 0
+    return context.position if context is not None else 0
 
 
 @method(function('count', nargs=1))
@@ -1081,6 +1035,8 @@ def evaluate(self, context=None):
     except TypeError:
         self.wrong_type("the second argument must be a string")
 
+    if index < 0:
+        return ''
     if self.symbol == 'substring-before':
         return arg1[:index]
     else:
@@ -1091,12 +1047,12 @@ def evaluate(self, context=None):
 # Boolean functions
 @method(function('boolean', nargs=1))
 def evaluate(self, context=None):
-    return self.boolean_value(self[0].get_results(context))
+    return self.boolean_value([x for x in self[0].select(context)])
 
 
 @method(function('not', nargs=1))
 def evaluate(self, context=None):
-    return not self.boolean_value(self[0].get_results(context))
+    return not self.boolean_value([x for x in self[0].select(context)])
 
 
 @method(function('true', nargs=0))
@@ -1196,9 +1152,9 @@ def evaluate(self, context=None):
     try:
         number = decimal.Decimal(arg)
         if number > 0:
-            return float(number.quantize(decimal.Decimal('1'), rounding='ROUND_HALF_UP'))
+            return number.quantize(decimal.Decimal('1'), rounding='ROUND_HALF_UP')
         else:
-            return float(round(number))
+            return round(number)
     except TypeError as err:
         self.wrong_type(str(err))
     except decimal.DecimalException as err:
