@@ -17,6 +17,7 @@ import locale
 import math
 import operator
 from collections.abc import MutableSequence
+from copy import copy
 from urllib.parse import urlparse
 
 from .exceptions import ElementPathError, ElementPathKeyError, \
@@ -26,7 +27,8 @@ from .namespaces import XSD_NAMESPACE, XML_NAMESPACE, XLINK_NAMESPACE, \
     XSD_ANY_ATOMIC_TYPE, get_namespace, qname_to_prefixed, prefixed_to_qname, \
     XSD_UNTYPED_ATOMIC
 from .datatypes import UntypedAtomic, XSD_BUILTIN_TYPES
-from .xpath_nodes import is_xpath_node
+from .xpath_nodes import is_xpath_node, is_attribute_node, is_element_node, \
+    is_document_node
 from .xpath1_parser import XPath1Parser
 from .xpath_context import XPathSchemaContext
 from .schema_proxy import AbstractSchemaProxy
@@ -168,6 +170,9 @@ class XPath2Parser(XPath1Parser):
     }
 
     PATH_STEP_LABELS = ('axis', 'function', 'kind test')
+    PATH_STEP_SYMBOLS = {
+        '(integer)', '(string)', '(float)', '(decimal)', '(name)', '*', '@', '..', '.', '(', '{'
+    }
 
     def __init__(self, namespaces=None, variables=None, strict=True, compatibility_mode=False,
                  default_namespace=None, function_namespace=None, schema=None, base_uri=None,
@@ -274,11 +279,14 @@ class XPath2Parser(XPath1Parser):
             return self
 
         def evaluate_(self, context=None):
-            item = self.get_argument(context)
-            if item is None:
+            arg = self.data_value(self.get_argument(context))
+            if arg is None:
                 return []
+
             try:
-                return self.cast(item)
+                if isinstance(arg, UntypedAtomic):
+                    return self.cast(arg.value)
+                return self.cast(arg)
             except ElementPathError as err:
                 if err.token is None:
                     err.token = self
@@ -288,7 +296,7 @@ class XPath2Parser(XPath1Parser):
             except TypeError as err:
                 raise self.error('FORG0006', str(err))
 
-        def cast(_value):
+        def cast(value):
             raise NotImplementedError
 
         pattern = r'\b%s(?=\s*\(|\s*\(\:.*\:\)\()' % symbol
@@ -371,7 +379,8 @@ class XPath2Parser(XPath1Parser):
                 pass
         else:
             # Static context evaluation with a dynamic schema context
-            for _ in root_token.select(context=self.schema.get_context()):
+            context = self.schema.get_context()
+            for _ in root_token.select(context):
                 pass
 
         return root_token
@@ -386,6 +395,7 @@ prefix = XPath2Parser.prefix
 infix = XPath2Parser.infix
 infixr = XPath2Parser.infixr
 method = XPath2Parser.method
+function = XPath2Parser.function
 
 ##
 # Remove symbols that have to be redefined for XPath 2.0.
@@ -413,14 +423,14 @@ XPath2Parser.duplicate('|', 'union')
 @method(infix('intersect', bp=55))
 def select(self, context=None):
     if context is not None:
-        results = set(self[0].select(context.copy())) & set(self[1].select(context.copy()))
+        results = set(self[0].select(copy(context))) & set(self[1].select(copy(context)))
         yield from context.iter_results(results)
 
 
 @method(infix('except', bp=55))
 def select(self, context=None):
     if context is not None:
-        results = set(self[0].select(context.copy())) - set(self[1].select(context.copy()))
+        results = set(self[0].select(copy(context))) - set(self[1].select(copy(context)))
         yield from context.iter_results(results)
 
 
@@ -440,7 +450,7 @@ def nud(self):
 
 @method('if')
 def evaluate(self, context=None):
-    if self.boolean_value(self[0].evaluate(context)):
+    if self.boolean_value(self[0].evaluate(copy(context))):
         return self[1].evaluate(context)
     else:
         return self[2].evaluate(context)
@@ -448,7 +458,7 @@ def evaluate(self, context=None):
 
 @method('if')
 def select(self, context=None):
-    if self.boolean_value([x for x in self[0].select(context)]):
+    if self.boolean_value([x for x in self[0].select(copy(context))]):
         yield from self[1].select(context)
     else:
         yield from self[2].select(context)
@@ -482,13 +492,13 @@ def evaluate(self, context=None):
         return
 
     some = self.symbol == 'some'
-    selectors = tuple(self[k].select(context.copy()) for k in range(1, len(self) - 1, 2))
+    selectors = tuple(self[k].select(copy(context)) for k in range(1, len(self) - 1, 2))
 
     for results in product(*selectors):
         for i in range(len(results)):
             context.variables[self[i * 2][0].value] = results[i]
 
-        if self.boolean_value([x for x in self[-1].select(context.copy())]):
+        if self.boolean_value([x for x in self[-1].select(copy(context))]):
             if some:
                 return True
         elif not some:
@@ -520,11 +530,11 @@ def nud(self):
 @method('for')
 def select(self, context=None):
     if context is not None:
-        selectors = tuple(self[k].select(context.copy()) for k in range(1, len(self) - 1, 2))
+        selectors = tuple(self[k].select(copy(context)) for k in range(1, len(self) - 1, 2))
         for results in product(*selectors):
             for i in range(len(results)):
                 context.variables[self[i * 2][0].value] = results[i]
-            yield from self[-1].select(context.copy())
+            yield from self[-1].select(copy(context))
 
 
 ###
@@ -656,10 +666,13 @@ def evaluate(self, context=None):
         else:
             self.wrong_context_type("an atomic value is required")
 
-    input_value = self.data_value(result[0])
+    arg = self.data_value(result[0])
+    if isinstance(arg, UntypedAtomic):
+        arg = arg.value
+
     try:
         if namespace != XSD_NAMESPACE:
-            value = self.parser.schema.cast_as(input_value, atomic_type)
+            value = self.parser.schema.cast_as(arg, atomic_type)
         else:
             local_name = atomic_type.split('}')[1]
             token_class = self.parser.symbol_table.get(local_name)
@@ -668,16 +681,16 @@ def evaluate(self, context=None):
                 self.unknown_atomic_type(msg % self[1].source)
 
             if local_name in {'base64Binary', 'hexBinary'}:
-                value = token_class.cast(input_value, self[0].label == 'literal')
+                value = token_class.cast(arg, self[0].label == 'literal')
             elif local_name in {'dateTime', 'date', 'gDay', 'gMonth',
                                 'gMonthDay', 'gYear', 'gYearMonth', 'time'}:
                 value = token_class.cast(
-                    input_value, tz=None if context is None else context.timezone
+                    arg, tz=None if context is None else context.timezone
                 )
             elif local_name == 'QName':
-                value = token_class.cast(input_value, self.parser.namespaces)
+                value = token_class.cast(arg, self.parser.namespaces)
             else:
-                value = token_class.cast(input_value)
+                value = token_class.cast(arg)
 
     except ElementPathError as err:
         if self.symbol != 'cast':
@@ -717,7 +730,7 @@ def evaluate(self, context=None):
 @method(',')
 def select(self, context=None):
     for op in self:
-        yield from op.select(context.copy() if context else None)
+        yield from op.select(context=copy(context))
 
 
 ###
@@ -758,14 +771,16 @@ def select(self, context=None):
 @method(infix('le', bp=30))
 @method(infix('ge', bp=30))
 def evaluate(self, context=None):
-    if context is None:
-        op1, op2 = self[0].get_atomized_operand(), self[1].get_atomized_operand()
-    else:
-        op1 = self[0].get_atomized_operand(context.copy())
-        op2 = self[1].get_atomized_operand(context.copy())
+    op1 = self[0].get_atomized_operand(context=copy(context))
+    op2 = self[1].get_atomized_operand(context=copy(context))
 
     if op1 is not None and op2 is not None:
-        return getattr(operator, self.symbol)(op1, op2)
+        try:
+            return getattr(operator, self.symbol)(op1, op2)
+        except TypeError as err:
+            if isinstance(context, XPathSchemaContext):
+                self.wrong_context_type(str(err))
+            self.wrong_type(str(err))
 
 
 ###
@@ -836,4 +851,186 @@ def evaluate(self, context=None):
     except decimal.DivisionByZero:
         raise self.error('FOAR0001')
 
+
+###
+# Kind tests (sequence types that can appear also in XPath expressions)
+@method(function('document-node', nargs=(0, 1), label='kind test'))
+def select(self, context=None):
+    if context is None:
+        self.missing_context()
+    elif not is_document_node(context.root) or context.item is not None:
+        return
+    elif not self:
+        for item in context.iter_children_or_self():
+            if item is None:
+                yield context.root
+    else:
+        context.item = context.root.getroot()
+        elements = [e for e in self[0].select(context) if is_element_node(e)]
+        if len(elements) == 1:
+            yield context.root
+        context.item = None
+
+
+@method('document-node')
+def nud(self):
+    self.parser.advance('(')
+    if self.parser.next_token.symbol == 'element':
+        self[0:] = self.parser.expression(5),
+        if self.parser.next_token.symbol == ',':
+            self.wrong_nargs('Too many arguments: expected at most 1 argument')
+    elif self.parser.next_token.symbol != ')':
+        self.parser.next_token.wrong_syntax('element() kind test expected')
+    self.parser.advance(')')
+    self.value = None
+    return self
+
+
+@method(function('element', nargs=(0, 2), label='kind test'))
+def select(self, context=None):
+    if context is None:
+        self.missing_context()
+    elif not self:
+        for item in context.iter_children_or_self():
+            if is_element_node(item):
+                yield item
+    else:
+        for item in self[0].select(context):
+            if len(self) == 1:
+                yield item
+            elif self.xsd_types:
+                type_annotation = self[1].evaluate(context)
+                if self.xsd_types.is_matching(type_annotation, self.parser.default_namespace):
+                    yield context.item
+
+
+@method('element')
+def nud(self):
+    self.parser.advance('(')
+    if self.parser.next_token.symbol != ')':
+        if self.parser.next_token.symbol not in {'(name)', ':', '*'}:
+            self.parser.next_token.wrong_syntax('a QName or a wildcard expected')
+        self[0:] = self.parser.expression(5),
+        if self.parser.next_token.symbol == ',':
+            self.parser.advance(',')
+            if self.parser.next_token.symbol not in {'(name)', ':'}:
+
+                self.parser.next_token.wrong_syntax('a QName expected')
+            self[1:] = self.parser.expression(5),
+    self.parser.advance(')')
+    self.value = None
+    return self
+
+
+@method(function('schema-attribute', nargs=1, label='kind test'))
+def select(self, context=None):
+    if context is not None:
+        for _ in context.iter_children_or_self():
+            attribute_name = self[0].source
+            qname = prefixed_to_qname(attribute_name, self.parser.namespaces)
+            if self.parser.schema.get_attribute(qname) is None:
+                self.missing_name("attribute %r not found in schema" % attribute_name)
+
+            if is_attribute_node(context.item, qname):
+                yield context.item
+
+
+@method(function('schema-element', nargs=1, label='kind test'))
+def select(self, context=None):
+    if context is not None:
+        for _ in context.iter_children_or_self():
+            element_name = self[0].source
+            qname = prefixed_to_qname(element_name, self.parser.namespaces)
+            if self.parser.schema.get_element(qname) is None \
+                    and self.parser.schema.get_substitution_group(qname) is None:
+                self.missing_name("element %r not found in schema" % element_name)
+
+            if is_element_node(context.item) and context.item.tag == qname:
+                yield context.item
+
+
+@method('schema-attribute')
+@method('schema-element')
+def nud(self):
+    self.parser.advance('(')
+    if self.parser.next_token.symbol not in ('(name)', ':'):
+        self.parser.next_token.wrong_syntax('a name expected')
+    self[0:] = self.parser.expression(5),
+    self.parser.advance(')')
+    self.value = None
+    return self
+
+
+###
+# Multi role-tokens definition: in XPath 2.0 the 'attribute' keyword is used both for
+# attribute:: axis and attribute() node type function.
+#
+# First the XPath1 token class has to be removed from the XPath2 symbol table. Then the
+# symbol has to be registered usually with the same binding power (bp --> lbp, rbp), a
+# multi-value label (using a tuple of values) and a custom pattern. Finally a custom nud
+# or led method is required.
+unregister('attribute')
+register('attribute', lbp=90, rbp=90, label=('kind test', 'axis'),
+         pattern=r'\battribute(?=\s*\:\:|\s*\(\:.*\:\)\s*\:\:|\s*\(|\s*\(\:.*\:\)\()')
+
+
+@method('attribute')
+def nud(self):
+    if self.parser.next_token.symbol == '::':
+        self.parser.advance('::')
+        self.parser.next_token.expected(
+            '(name)', '*', 'text', 'node', 'document-node', 'comment', 'processing-instruction',
+            'attribute', 'schema-attribute', 'element', 'schema-element'
+        )
+        self[:] = self.parser.expression(rbp=90),
+        self.label = 'axis'
+    else:
+        self.parser.advance('(')
+        if self.parser.next_token.symbol != ')':
+            self[:] = self.parser.expression(5),
+            if self.parser.next_token.symbol == ',':
+                self.parser.advance(',')
+                self[1:] = self.parser.expression(5),
+        self.parser.advance(')')
+        self.label = 'kind test'
+    return self
+
+
+@method('attribute')
+def select(self, context=None):
+    if context is None:
+        return
+    elif self.label == 'axis':
+        for _ in context.iter_attributes():
+            yield from self[0].select(context)
+    elif not self:
+        for item in context.iter_attributes():
+            if is_attribute_node(item):
+                yield context.item[1]
+    else:
+        name = self[0].value
+        if self.parser.schema is not None and len(self) == 2:
+            type_name = prefixed_to_qname(self[1].value, namespaces=self.parser.namespaces)
+        else:
+            type_name = None
+
+        for item in context.iter_attributes():
+            if is_attribute_node(item, name):
+                if isinstance(context, XPathSchemaContext):
+                    self.add_xsd_type(item[0], item[1].type)
+                elif not type_name:
+                    yield context.item[1]
+                else:
+                    xsd_type = self.get_xsd_type(item)
+                    if xsd_type is not None and xsd_type.name == type_name:
+                        yield context.item[1]
+
+
+@method('attribute')
+def evaluate(self, context=None):
+    if context is not None:
+        if is_attribute_node(context.item, self[0].evaluate(context) if self else None):
+            return context.item[1]
+
+
 # XPath 2.0 definitions continue into module xpath2_functions
