@@ -12,7 +12,8 @@ from functools import lru_cache
 
 from .exceptions import ElementPathTypeError
 from .xpath_nodes import AttributeNode, TextNode, TypedAttribute, TypedElement, \
-    is_etree_element, is_element_node, is_document_node, is_attribute_node
+    etree_iter_nodes, is_etree_element, is_element_node, is_document_node, \
+    is_attribute_node
 
 
 class XPathContext(object):
@@ -35,7 +36,7 @@ class XPathContext(object):
     :param timezone: implicit timezone to be used when a date, time, or dateTime value does \
     not have a timezone.
     """
-    def __init__(self, root, item=None, position=0, size=1, axis=None,
+    def __init__(self, root, item=None, position=1, size=1, axis=None,
                  variables=None, current_dt=None, timezone=None,
                  documents=None, collections=None, default_collection=None):
         if not is_element_node(root) and not is_document_node(root):
@@ -66,13 +67,13 @@ class XPathContext(object):
             self.__class__.__name__, self.root, self.item, self.position, self.size, self.axis
         )
 
-    def copy(self, clear_axis=True):
+    def __copy__(self):
         obj = type(self)(
             root=self.root,
             item=self.item,
             position=self.position,
             size=self.size,
-            axis=None if clear_axis else self.axis,
+            axis=None,
             variables=self.variables.copy(),
             current_dt=self.current_dt,
             timezone=self.timezone,
@@ -86,6 +87,14 @@ class XPathContext(object):
         obj._parent_map = self._parent_map
         return obj
 
+    def copy(self, clear_axis=True):
+        if clear_axis:
+            return self.__copy__()
+        else:
+            obj = self.__copy__()
+            obj.axis = self.axis
+            return obj
+
     @property
     def parent_map(self):
         if self._parent_map is None:
@@ -138,17 +147,71 @@ class XPathContext(object):
         else:
             return is_element_node(self.item)
 
-    # Context item iterators
+    def iter(self):
+        """Iterates context nodes, including text and attribute nodes."""
+        root = self.root
+        if is_document_node(root):
+            yield root
+            root = root.getroot()
+        yield from etree_iter_nodes(root, with_attributes=True)
+
+    def iter_results(self, results):
+        """
+        Iterates results in document order.
+
+        :param results: list containing selection results.
+        """
+        status = self.item, self.size, self.position
+        root = self.root.getroot() if is_document_node(self.root) else self.root
+
+        self.size = len(results)
+        for self.position, self.item in \
+                enumerate(etree_iter_nodes(root, with_attributes=True), start=1):
+            if self.item in results:
+                yield self.item
+            elif isinstance(self.item, AttributeNode):
+                # Match XSD decoded attributes
+                for attr in filter(lambda x: isinstance(x, TypedAttribute), results):
+                    if attr[0] == self.item:
+                        yield attr
+            elif is_etree_element(self.item):
+                # Match XSD decoded elements
+                for elem in filter(lambda x: isinstance(x, TypedElement), results):
+                    if elem[0] is self.item:
+                        yield elem
+
+        self.item, self.size, self.position = status
+
+    def iter_selector(self, selector):
+        """
+        Iterator for generic selector.
+
+        :param selector: a selector generator function.
+        """
+        status = self.item, self.size, self.position
+
+        results = [x for x in selector(self.copy())]
+        self.size = len(results)
+        for self.position, self.item in enumerate(results, start=1):
+            yield self.item
+
+        self.item, self.size, self.position = status
+
+    ##
+    # Context item iterators for axis
+
     def iter_self(self):
-        status = self.item, self.size, self.position, self.axis
+        """Iterator for 'self' axis and '.' shortcut."""
+        status = self.size, self.position, self.axis
+        self.size = self.position = 1
         self.axis = 'self'
         yield self.item
-        self.item, self.size, self.position, self.axis = status
+        self.size, self.position, self.axis = status
 
     def iter_attributes(self):
+        """Iterator for 'attribute' axis and '@' shortcut."""
         if not is_element_node(self.item):
             return
-        self._elem = self.item
 
         status = self.item, self.size, self.position, self.axis
         self.axis = 'attribute'
@@ -156,12 +219,20 @@ class XPathContext(object):
         if isinstance(self.item, TypedElement):
             self.item = self.item.elem
 
-        for self.item in map(lambda x: AttributeNode(*x), self.item.attrib.items()):
+        attributes = [AttributeNode(*x) for x in self.item.attrib.items()]
+
+        self.size = len(attributes)
+        for self.position, self.item in enumerate(attributes, start=1):
             yield self.item
 
         self.item, self.size, self.position, self.axis = status
 
-    def iter_children_or_self(self, item=None, child_axis=False):
+    def iter_children_or_self(self, child_axis=False):
+        """
+        Iterator for 'child' forward axis and '/' step.
+
+        :param child_axis: if set to `True` use the child axis anyway.
+        """
         if not child_axis and self.axis is not None:
             yield self.item
             return
@@ -169,13 +240,11 @@ class XPathContext(object):
         status = self.item, self.size, self.position, self.axis
         self.axis = 'child'
 
-        if item is not None:
-            self.item = item[0] if isinstance(item, TypedElement) else item
-        elif isinstance(self.item, TypedElement):
+        if isinstance(self.item, TypedElement):
             self.item = self.item[0]
 
         if self.item is None:
-            self.size, self.position = 1, 0
+            self.size = self.position = 1
             self.item = self.root.getroot() if is_document_node(self.root) else self.root
             yield self.item
         elif is_etree_element(self.item):
@@ -184,14 +253,135 @@ class XPathContext(object):
                 self.item = TextNode(elem.text)
                 yield self.item
             self.size = len(elem)
-            for self.position, self.item in enumerate(elem):
+            for self.position, self.item in enumerate(elem, start=1):
                 yield self.item
 
         self.item, self.size, self.position, self.axis = status
 
+    def iter_parent(self):
+        """Iterator for 'parent' reverse axis and '..' shortcut."""
+        if isinstance(self.item, TypedElement):
+            parent = self.get_parent(self.item[0])
+        else:
+            parent = self.get_parent(self.item)
+
+        if parent is not None:
+            status = self.item, self.size, self.position, self.axis
+            self.axis = 'parent'
+
+            self.item = parent
+            self.size = self.position = 1
+            yield self.item
+
+            self.item, self.size, self.position, self.axis = status
+
+    def iter_siblings(self, axis=None):
+        """
+        Iterator for 'following-sibling' forward axis and 'preceding-sibling' reverse axis.
+
+        :param axis: the context axis, default is 'following-sibling'.
+        """
+        if isinstance(self.item, TypedElement):
+            item = self.item[0]
+        elif not is_etree_element(self.item) or callable(self.item.tag):
+            return
+        else:
+            item = self.item
+
+        parent = self.get_parent(item)
+        if parent is None:
+            return
+
+        status = self.item, self.size, self.position, self.axis
+        self.axis = axis or 'following-sibling'
+
+        siblings = []
+        if axis == 'preceding-sibling':
+            for child in parent:  # pragma: no cover
+                if child is item:
+                    break
+                siblings.append(child)
+
+            self.size = self.position = len(siblings)
+            for self.item in siblings:
+                yield self.item
+                self.position -= 1
+        else:
+            follows = False
+            for child in parent:
+                if follows:
+                    siblings.append(child)
+                elif child is item:
+                    follows = True
+
+            self.size = len(siblings)
+            for self.position, self.item in enumerate(siblings, start=1):
+                yield self.item
+
+        self.item, self.size, self.position, self.axis = status
+
+    def iter_descendants(self, item=None, axis=None):
+        """
+        Iterator for 'descendant' and 'descendant-or-self' forward axes and '//' shortcut.
+
+        :param item: use another item instead of the context's item.
+        :param axis: the context axis, for default has no explicit axis.
+        """
+        status = self.item, self.size, self.position, self.axis
+        self.axis = axis
+
+        if item is not None:
+            self.item = item[0] if isinstance(item, TypedElement) else item
+        elif isinstance(self.item, TypedElement):
+            self.item = self.item[0]
+
+        if self.item is None:
+            self.size = self.position = 1
+            yield self.root
+            self.item = self.root.getroot() if is_document_node(self.root) else self.root
+        elif not is_element_node(self.item):
+            return
+
+        if axis == 'descendant':
+            descendants = [x for x in etree_iter_nodes(self.item, with_root=False)]
+        else:
+            descendants = [x for x in etree_iter_nodes(self.item)]
+
+        self.size = len(descendants)
+        for self.position, self.item in enumerate(descendants, start=1):
+            yield self.item
+
+        self.item, self.size, self.position, self.axis = status
+
+    def iter_ancestors(self, axis=None):
+        """
+        Iterator for 'ancestor-or-self' and 'ancestor' reverse axes.
+
+        :param axis: the context axis, default is 'ancestor-or-self'.
+        """
+        status = self.item, self.size, self.position, self.axis
+        self.axis = axis or 'ancestor-or-self'
+
+        if isinstance(self.item, TypedElement):
+            self.item = self.item[0]
+
+        ancestors = [self.item] if axis == 'ancestor-or-self' else []
+        parent = self.get_parent(self.item)
+        while parent is not None:
+            ancestors.append(parent)
+            parent = self.get_parent(parent)
+
+        self.size = self.position = len(ancestors) or 1
+        for self.item in reversed(ancestors):
+            yield self.item
+            self.position -= 1
+
+        self.item, self.size, self.position, self.axis = status
+
     def iter_preceding(self):
+        """Iterator for 'preceding' reverse axis."""
         item = self.item[0] if isinstance(self.item, TypedElement) else self.item
-        if not is_etree_element(item):
+        if not is_etree_element(item) or item is self.root or callable(item.tag):
             return
 
         status = self.item, self.size, self.position, self.axis
@@ -207,128 +397,47 @@ class XPathContext(object):
                 ancestors.append(parent)
                 elem = parent
 
-        for elem in self.root.iter():
+        preceding = []
+        for elem in self.root.iter():  # pragma: no cover
             if elem is item:
                 break
             elif elem not in ancestors:
-                self.item = elem
-                yield elem
+                preceding.append(elem)
+
+        self.size = len(preceding)
+        for self.position, self.item in enumerate(preceding, start=1):
+            yield self.item
 
         self.item, self.size, self.position, self.axis = status
 
-    def iter_parent(self, axis=None):
-        if isinstance(self.item, TypedElement):
-            parent = self.get_parent(self.item[0])
-        else:
-            parent = self.get_parent(self.item)
-
-        if parent is not None:
-            status = self.item, self.size, self.position, self.axis
-            self.axis = axis
-
-            self.item = parent
-            yield self.item
-
-            self.item, self.size, self.position, self.axis = status
-
-    def iter_descendants(self, item=None, axis=None):
+    def iter_followings(self):
+        """Iterator for 'following' forward axis."""
         status = self.item, self.size, self.position, self.axis
-        self.axis = axis
+        self.axis = 'following'
 
-        if item is not None:
-            self.item = item[0] if isinstance(item, TypedElement) else item
+        if self.item is None or self.item is self.root:
+            return
         elif isinstance(self.item, TypedElement):
             self.item = self.item[0]
-
-        if self.item is None:
-            self.size, self.position = 1, 0
-            yield self.root
-            self.item = self.root.getroot() if is_document_node(self.root) else self.root
-        elif not is_element_node(self.item):
+        elif not is_etree_element(self.item) or callable(self.item.tag):
             return
 
-        yield from self._iter_descendants()
+        descendants = set(etree_iter_nodes(self.item))
+        followings = []
+        follows = False
+        for elem in etree_iter_nodes(self.root):
+            if follows:
+                if elem not in descendants:
+                    followings.append(elem)
+            elif self.item is elem:
+                follows = True
+
+        self.size = len(followings)
+        for self.position, self.item in enumerate(followings, start=1):
+            yield self.item
 
         self.item, self.size, self.position, self.axis = status
 
-    def _iter_descendants(self):
-        elem = self._elem = self.item
-        yield elem
-        if elem.text is not None:
-            self.item = TextNode(elem.text)
-            yield self.item
-        if len(elem):
-            self.size = len(elem)
-            for self.position, self.item in enumerate(elem):
-                yield from self._iter_descendants()
-
-    def iter_ancestors(self, item=None, axis=None):
-        status = self.item, self.size, self.position, self.axis
-        self.axis = axis
-
-        if item is not None:
-            self.item = item[0] if isinstance(item, TypedElement) else item
-        elif isinstance(self.item, TypedElement):
-            self.item = self.item[0]
-
-        while True:
-            parent = self.get_parent(self.item)
-            if parent is None:
-                break
-            else:
-                self.item = parent
-                yield parent
-
-        self.item, self.size, self.position, self.axis = status
-
-    def iter(self):
-        """Iterates context nodes."""
-        status = self.item, self.size, self.position
-        self.item = self.root
-        if is_document_node(self.item):
-            yield self.item
-            self.item = self.item.getroot()
-
-        yield from self._iter_context()
-
-        self.item, self.size, self.position, = status
-
-    def iter_results(self, results):
-        """Iterates results in document order."""
-        status = self.item, self.size, self.position
-        self.item = self.root.getroot() if is_document_node(self.root) else self.root
-
-        for item in self._iter_context():
-            if item in results:
-                yield item
-            elif isinstance(item, AttributeNode):
-                # Match XSD decoded attributes
-                for attr in filter(lambda x: isinstance(x, TypedAttribute), results):
-                    if attr[0] == item:
-                        yield attr
-            elif is_etree_element(item):
-                # Match XSD decoded elements
-                for elem in filter(lambda x: isinstance(x, TypedElement), results):
-                    if elem[0] is item:
-                        yield elem
-
-        self.item, self.size, self.position = status
-
-    def _iter_context(self):
-        elem = self._elem = self.item
-        yield elem
-        if elem.text is not None:
-            self.item = TextNode(elem.text)
-            yield self.item
-
-        for self.item in map(lambda x: AttributeNode(*x), elem.attrib.items()):
-            yield self.item
-
-        if len(elem):
-            self.size = len(elem)
-            for self.position, self.item in enumerate(elem):
-                yield from self._iter_context()
-
 
 class XPathSchemaContext(XPathContext):
     """
